package se.gawell.setrace.spring;

import java.util.List;

import net.sf.cglib.proxy.Callback;
import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.NoOp;

import org.springframework.aop.TargetSource;
import org.springframework.aop.framework.autoproxy.AbstractAutoProxyCreator;
import org.springframework.util.PatternMatchUtils;

import se.gawell.setrace.logging.Logger;
import se.gawell.setrace.logging.LoggerFactory;

/**
 * Auto proxy creator for creating proxies used for logging method invocations.
 * 
 * Use this as an AbstractAutoProxyCreator.
 * 
 * @author Niklas Gawell
 *
 */
public class ExcludeMappedNameBeanNameAutoProxyCreator extends
		AbstractAutoProxyCreator {

	private final LoggerFactory loggerFactory;
	private final List<String> excludedNames;

	public ExcludeMappedNameBeanNameAutoProxyCreator(LoggerFactory loggerFactory, List<String> excludedNames) {
		super();
		this.loggerFactory = loggerFactory;
		this.excludedNames = excludedNames;
	}

	@Override
	protected Object[] getAdvicesAndAdvisorsForBean(Class<?> beanClass,
			String beanName, TargetSource targetSource) {
		if (classShouldBeExcludedSinceItIsNotProxable(beanClass)) {
			return DO_NOT_PROXY;
		}
		for (String nameToExclude : excludedNames) {
			if (beanNameMatchExclude(beanClass.getName(), nameToExclude)) {
				return DO_NOT_PROXY;
			}
		}
		logProxyCreation(beanName);
		return PROXY_WITHOUT_ADDITIONAL_INTERCEPTORS;
	}

	private void logProxyCreation(String beanName) {
		Logger loggerForClass = loggerFactory.getLoggerForClass(getClass());
		loggerForClass.debug("Adding method execution logger to bean {0}.", beanName);
	}

	private boolean classShouldBeExcludedSinceItIsNotProxable(Class<?> beanClass) {
		try {
			Callback callback = NoOp.INSTANCE;
			Enhancer.create(beanClass, callback );
		} catch (RuntimeException e) {
			return true;
		}
		return false;
	}

	private boolean beanNameMatchExclude(String beanName, String nameToExclude) {
		return PatternMatchUtils.simpleMatch(nameToExclude, beanName);
	}

}
