package se.gawell.setrace.spring;

import java.lang.reflect.Method;
import java.util.List;

import org.springframework.aop.AfterReturningAdvice;
import org.springframework.aop.MethodBeforeAdvice;
import org.springframework.aop.ThrowsAdvice;

import se.gawell.setrace.logging.Logger;
import se.gawell.setrace.logging.LoggerFactory;
import se.gawell.setrace.logging.ToStringGenerator;

/**
 * Advices to log before method, after method returning or after method throws
 * exception.
 * 
 * @author Niklas Gawell
 * 
 */
public class MethodExecutionLogger implements MethodBeforeAdvice, AfterReturningAdvice, ThrowsAdvice {

	private final LoggerFactory loggerFactory;
	private final List<String> methodNamesToExclude;
	private final ToStringGenerator toStringGenerator;

	protected MethodExecutionLogger(LoggerFactory loggerFactory, List<String> methodNamesToExclude,
			ToStringGenerator toStringGenerator) {
		super();
		this.loggerFactory = loggerFactory;
		this.methodNamesToExclude = methodNamesToExclude;
		this.toStringGenerator = toStringGenerator;
	}

	@Override
	public void before(Method method, Object[] args, Object target) throws Throwable {
		if (methodShouldBeIgnored(method)) {
			return;
		}
		Logger logger = getLoggerForTarget(target);
		String debugMessage = generateMethodNameWithParametersDebugMessage(method, args);
		logger.debug(debugMessage, argumentListToString(args));
	}

	@Override
	public void afterReturning(Object returnValue, Method method, Object[] args, Object target) throws Throwable {
		if (methodShouldBeIgnored(method)) {
			return;
		}
		Logger logger = getLoggerForTarget(target);
		String parameterListAsString = generateParameterList(args);
		String debugMessage = generateReturnDebugMessage(returnValue, method, parameterListAsString);
		logger.debug(debugMessage, argumentListToString(args));
	}

	private Object[] argumentListToString(Object[] args) {
		String[] strings = new String[args.length];
		for (int i = 0; i < args.length; i++) {
			strings[i] = reflectionToString(args[i]);
		}
		return strings;
	}

	public void afterThrowing(Method method, Object[] args, Object target, Throwable t) throws Throwable {
		if (methodShouldBeIgnored(method)) {
			return;
		}
		Logger logger = getLoggerForTarget(target);
		String debugMessage = generateThrewDebugMessage(method, args, t);
		logger.debug(debugMessage, argumentListToString(args));
	}

	private String generateThrewDebugMessage(Method method, Object[] args, Throwable t) {
		String methodNameWithArguments = generateMethodNameWithParametersDebugMessage(method, args);
		StringBuilder builder = new StringBuilder();
		builder.append(methodNameWithArguments).append(" threw ").append(t.toString());
		String debugMessage = builder.toString();
		return debugMessage;
	}

	private String generateMethodNameWithParametersDebugMessage(Method method, Object[] args) {
		String parameterListAsString = generateParameterList(args);
		String methodNameWithArguments = generateStringForMethodWithArguments(method, parameterListAsString);
		return methodNameWithArguments;
	}

	private boolean methodShouldBeIgnored(Method method) {
		for (String methodNameToExclude : methodNamesToExclude) {
			if (method.getName().equals(methodNameToExclude)) {
				return true;
			}
		}
		return false;
	}

	private Logger getLoggerForTarget(Object target) {
		if (target != null) {
			return loggerFactory.getLoggerForClass(target.getClass());
		} else {
			return loggerFactory.getLoggerForClass(null);
		}
	}

	private String generateReturnDebugMessage(Object returnValue, Method method, String parameterListAsString) {
		StringBuilder returnStringBuilder = new StringBuilder();
		returnStringBuilder.append(generateStringForMethodWithArguments(method, parameterListAsString));
		returnStringBuilder.append(" returned ");
		returnStringBuilder.append(reflectionToString(returnValue));

		returnStringBuilder.append(" (").append(method.getReturnType().getSimpleName()).append(")");
		return returnStringBuilder.toString();
	}

	private String generateStringForMethodWithArguments(Method method, String parameterListAsString) {
		return method.getName() + "(" + parameterListAsString + ")";
	}

	private String generateParameterList(Object[] args) {
		StringBuilder parameterList = new StringBuilder();
		for (int i = 0; i < args.length; i++) {
			if (i > 0) {
				parameterList.append(", ");
			}
			parameterList.append("{").append(i).append("}");
		}
		String parameterListAsString = parameterList.toString();
		return parameterListAsString;
	}

	public String reflectionToString(Object object) {
		StringBuilder stringBuilder = new StringBuilder();
		toStringGenerator.generateToString(stringBuilder, object);
		return stringBuilder.toString();
	}

}
